<!--	
	Library of routines to perform periodic backups and table maintenance;
	
	Implements:
		_Maintenace_Schedule;	(private routine for Lasso Administration)

	Variables:
		__Maintenance_Status__ is Active/Sleep/Pause/Abort
			Active - Check schedule queue within one minute, sleep until next event 
			Sleep - Sleep until indicated time
			Pause - Pause indefinitely
			Abort - Kill the session watcher
		__Maintenance_Sleep__ is date/time at which to wake up

	History:
		10/30/2003	JS	- added special handling of delays that are multiples of 1440
	    10/1/2003   FMS - LP7 syntax update, privileged tag, new status variable, silent sleep.
	    08/13/2002  FMS - Created based on scheduleLibrary.lasso
-->


<!--------------------------------------------------------------------->
<!-- _Maintenance_Schedule:
		Queues a Maintenance item for later execution.  Roughly equivalent to old Post_Inline;

	Inputs:
		-Start=(Date),			//	defaults to today
		-End=(Date),			//	defaults to never
		-URL=(ExecuteURL),		//	URL to execute (simulate a browser hit)
	    -Repeat=(true/false),	//	Should this repeat? Default to true if -Delay passed
	    -Restart=(true/false),	//	Continue executing after restart.  Defaults to true
	    -Delay=(#Minutes),
	    -Username,				//	optional username for authentication
	    -Password;				//	optional password for authentication

	Outputs:
		None

	Side-Effects:
		Appends a record to site._maintenance that contains all info required to execute the item
-->
<?LassoScript

Define_Tag: '_Maintenance_Schedule', -Priority='replace',
		-Required='event',
		-Optional='start',
		-Optional='end',
		-Optional='delay',
		-Optional='repeat',
		-Optional='restart';

	// Create an array to store the constructed event
	local: 'event_array' = (array);

	// Insert the event
	local: 'l_event' = (string: (local: 'event'));
	#l_event->trim;
	#event_array->(insert: 'executeAction' = (local: 'l_event'));

	local: 'action_array' = #l_event->(split: ':');
	if: (#action_array->type == 'array') && (#action_array->size >= 3);
		local: 'event_display' =  #action_array->(get:1) + ' Database `' + #action_array->(get:2) + '` Tables ' + #action_array->(get:3);
	else;
		local: 'event_display' =  #l_event;
	/if;
	log_warning: 'Maintenance: Queue Action ' + #event_display;

	// If no start date defined, then default to today
	if: ((local: 'start') == '');
		#event_array->(insert: 'date_time_start' = (Date)->(Format: '%Q %T'));
		#event_array->(insert: 'next_execute' = (Date)->(Format: '%Q %T'));
	else;
		#event_array->(insert: 'date_time_start' = (Date: (string: (local: 'start')))->(Format: '%Q %T'));
		#event_array->(insert: 'next_execute' = (Date: (string: (local: 'start')))->(Format: '%Q %T'));
	/If;

	// If no end date defined, then default to never
	if: ((local: 'end') == '');
		#event_array->(insert: 'date_time_end' = '3001-01-01 00:00:00');
	else;
		#event_array->(insert: 'date_time_end' = (Date: (string: (local: 'end')))->(Format: '%Q %T'));
	/if;

	// If no delay defined, then default to 0
	if: ((local: 'delay') == '');
		local: 'l_delay' = 0;
	else: ((integer: (local: 'delay')) < 0);
		local: 'l_delay' = 0;
	else;
		local: 'l_delay' = (integer: (local: 'delay'));
	/if;
	#event_array->(insert: 'delay'=#l_delay);

	// If no repeat defined, then default to true iff a delay is passed
	if: !(local_defined: 'repeat');
		if: (#l_delay > 0);
			#event_array->(insert: 'repeat'='1');
		else;
			#event_array->(insert: 'repeat'='0');
		/if;
	else: (#l_delay == 0);
		#event_array->(insert: 'repeat'='0');
	else;
		#event_array->(insert: 'repeat'=(integer: (boolean: (local: 'repeat'))));
	/if;

	// If no restart defined, then default to true
	if: !(local_defined: 'restart');
		#event_array->(insert: 'restart'='1');
	else;
		#event_array->(insert: 'restart'=(integer: (boolean: (local: 'restart'))));
	/if;

	// Add the event to the database
	inline: -database='site', -table='_maintenance', -add, #event_array;
		if: (error_currenterror) != (error_noerror);
			log_warning: 'Maintenance: ' + (error_currenterror) + ' (' + (error_currenterror: -errorcode) + ')';
		/if;
	/inline;
	
	// After adding an event, check the event queue immediately
	Global: '__Maintenance_Status__' = 'Active';

/Define_Tag;

?>

<!-- 
	Maintenance Initialization Code

	Actions:
		Creates new _maintenance table if it doesn't exist already
		Creates global __Maintenance_Status__ variable
		Creates global __Maintenance_Sleep__ variable
		Deletes events that are not set to restart
		Starts watcher and loops waiting for events to be scheduled

	Note:
		This code is only designed to be called once from LassoStartup
		It is formatted as a compound expression so Local variables
		can be used.

-->

<?LassoScript
	// This startup script is enclosed in a compound expression so the variables will not
	// show up in the list of global variables.
	{
		Log_Warning: 'Maintenance: Initializing Event Queue';
		
		var: '__Maintenance_Status__' = 'Active';
		var: '__Maintenance_Sleep__' = (Date);
		
		inline: -database='lasso_internal', -table='global_prefs', -search, -op='eq', 'store_key'='lasso_schedule_pause';
			if: (field: 'data') == 'Y';
				var: '__Maintenance_Status__'='Pause';
			else;
				var: '__Maintenance_Status__'='Active';
			/if;
		/inline;
	
		// Check if the table exists
		Local: 'table_exists' = false;
		local: 'table_fields' = (array);
		Inline: -Database='site', -SQL='SHOW COLUMNS FROM _maintenance';
			if: (found_count) > 0;
				#table_exists = true;
				records;
					#table_fields->(insert: (field: 'field')=(field: 'type'));
				/records;
			/if;
		/inline;
		
		if: #table_exists == false;
			// Create table if it doesn't exist already
			Inline: -Database='site',
				-SQL='CREATE TABLE IF NOT EXISTS _maintenance ' +
				'(' +
					'id INT PRIMARY KEY AUTO_INCREMENT,' +
					'date_time_start DATETIME,' +
					'date_time_end DATETIME,' +
					'next_execute DATETIME,' +
					'executeAction TEXT,' +
					'repeat TINYINT,' +
					'restart TINYINT,' +
					'delay INT,' +
					'user VARCHAR(32),' +
					'pass VARCHAR(32),' +
					'INDEX (date_time_end),' +
					'INDEX (next_execute)' +
				')';
			/Inline;
		/if;
	
		// Delete all records that are marked as non-restart or are past end date
		Inline: -Database='site',
			-SQL='DELETE FROM _maintenance WHERE restart=0 OR date_time_end <= NOW() OR next_execute = \'\'';
		/Inline;
	
		// Push the next_execute date to the next even multiple
		// This prevents events from ever executing twice rapidly
		Inline: -Database='site', -SQL='SELECT * FROM _maintenance WHERE date_time_end >= NOW() AND next_execute <= NOW() ORDER BY next_execute';
			Records;
				local: 'next_execute' = (Date: (Field: 'next_execute'), -format='%Q %T');
				local: 'date_fix' = (Date)->(Difference: (String: (Date)), -MinutesBetween);
				local: 'next_difference' = #next_execute->(Difference: (Date), -MinutesBetween) + #date_fix;
				
				local: 'delay' = (integer: (field: 'delay'));
	
				If: (#delay > 0);
					local: 'next_difference' = #next_difference + #delay - (#next_difference % #delay);
					#next_execute->(add: -minute=#next_difference);
					
					Inline:  -Database='site', -SQL='UPDATE _maintenance SET next_execute=\"' + (Date_Format: #next_execute, -format='%Q %T') + '\" WHERE id = ' + (Field: 'id');
					/Inline;
				/If;
			/Records;
		/Inline;
		
	}->(Run);

	// The rest of the script is run asynchronously

	{	
		// 
		// Start Maintenance Watcher
		// 
		Log_Warning: 'Maintenance: Starting Event Queue';	
		
		// 15 Second Delay at Startup
		Sleep: 15000; 
		
		While: ((Global: '__Maintenance_Status__') != 'Abort');
			Protect;
				If: ((Global: '__Maintenance_Status__') == 'Pause') || ((Global: '__Maintenance_Status__') == 'Stop');
					local: 'execute' = false;
				else: ((Global: '__Maintenance_Status__') == 'Sleep');
					if: (Global: '__Maintenance_Sleep__')->(isa: 'date') && ((Global: '__Maintenance_Sleep__') <= Date);
						local: 'execute' = true;
					else;
						local: 'execute' = false;
					/if;
				else;
					local: 'execute' = true;
				/if;
				
				If: (#execute == true);
				
					//
					// Check Event Queue
					//
				
					Log_Warning: 'Maintenance: Checking Event Queue';
						
					// Delete all records that are past end date.
					Inline: -Database='site', -SQL='DELETE FROM _maintenance WHERE date_time_end <= NOW()';
					/Inline;
				
					// Set the sleep time to one hour
					Global: '__Maintenance_Sleep__' = (Date);
					(Global: '__Maintenance_Sleep__')->(Add: -Hour=1);
				
					// Execute all records that are not past end date and which have next_execute earlier than now.
					Inline: -Database='site', -SQL='SELECT * FROM _maintenance WHERE date_time_end >= NOW() AND next_execute <= NOW() ORDER BY next_execute';
						Records;
							If: ((Global: '__Maintenance_Status__') == 'Pause') || ((Global: '__Maintenance_Status__') == 'Stop');
								Loop_Abort;
							/If;
							
							local: 'executeAction' = (field: 'executeAction');
							#executeAction->(trim);
				
							local: 'action_array' = #executeAction->(split: ':');
							if: (#action_array->type == 'array') && (#action_array->size == 4);
								local: 'event_display' =  #action_array->(get:1) + ' Database `' + #action_array->(get:2) + ' (' + #action_array->(get:3) + ')` Tables ' + #action_array->(get:4);
							else: (#action_array->type == 'array') && (#action_array->size == 3);
								local: 'event_display' =  #action_array->(get:1) + ' Database `' + #action_array->(get:2) + '` Tables ' + #action_array->(get:3);
							else;
								local: 'event_display' =  'Error: Invalid Action (' + #executeAction + ')';
							/if;
			
							log_warning: 'Maintenance: Executing Action ' + #event_display;
				
							Protect;
							
								if: #action_array->size == 3;
									local: 'action' = #action_array->(get:1);
									local: 'database' = #action_array->(get:2);
									local: 'alias' = #database;
									local: 'tables' = #action_array->(get:3);
								else: #action_array->size == 4;
									local: 'action' = #action_array->(get:1);
									local: 'database' = #action_array->(get:2);
									local: 'alias' = #action_array->(get:3);
									local: 'tables' = #action_array->(get:4);
								else;
									log_warning: 'Maintenance: Error: No action was specified';
									return;
								/if;
							
								if: #action >> 'flush';
									local: 'sql'='FLUSH TABLE ' + #tables;
								else: #action >> 'analyze';
									local: 'sql'='ANALYZE TABLE ' + #tables;
								else: #action >> 'optimize';
									local: 'sql'='OPTIMIZE TABLE ' + #tables;
								else: #action >> 'check quick';
									local: 'sql'='CHECK TABLE ' + #tables + ' QUICK';
								else: #action >> 'check medium';
									local: 'sql'='CHECK TABLE ' + #tables + ' MEDIUM';
								else: #action >> 'check extended';
									local: 'sql'='CHECK TABLE ' + #tables + ' EXTENDED';
								else: #action >> 'repair quick';
									local: 'sql'='REPAIR TABLE ' + #tables + ' QUICK';
								else: #action >> 'repair medium';
									local: 'sql'='REPAIR TABLE ' + #tables + ' MEDIUM';
								else: #action >> 'repair extended';
									local: 'sql'='REPAIR TABLE ' + #tables + ' EXTENDED';
								else: #action >> 'backup';
							
									local: 'export_folder' = (_admin_servicepath) + 'Admin/Backup/';
							
									if: !(file_exists: #export_folder) && !(file_exists: (string_removetrailing: #export_folder, -pattern='/'));
										file_create: #export_folder;
									/if;
									if: !(file_exists: #export_folder) && !(file_exists: (string_removetrailing: #export_folder, -pattern='/'));
										log_critical: 'Maintenance: Error: Backup folder not found ' + #export_folder;
										return;
									/if;
							
									local: 'database_folder' = #export_folder + #database + '/';
									if: !(file_exists: #database_folder) && !(file_exists: (string_removetrailing: #database_folder, -pattern='/'));
										file_create: #database_folder;
									/if;
									if: !(file_exists: #database_folder) && !(file_exists: (string_removetrailing: #database_folder, -pattern='/'));
										log_critical: 'Maintenance: Error: Could not create export folder ' + #database_folder;
										return;
									/if;
							
									local: 'datetime_folder' = #database_folder + (date->format: '%Y%m%d_%H%M%S') + '/';
									if: !(file_exists: #datetime_folder) && !(file_exists: (string_removetrailing: #datetime_folder, -pattern='/'));
										file_create: #datetime_folder;
									/if;
									if: !(file_exists: #datetime_folder) && !(file_exists: (string_removetrailing: #datetime_folder, -pattern='/'));
										log_critical: 'Maintenance: Error: Could not create export folder ' + #datetime_folder;
										return;
									/if;
							
									if: (lasso_version: -lassoplatform) >> 'Win32';
										local: 'temp_folder' = (string_replace: -find='/', -replace='\\\\', #datetime_folder);
									else;
										local: 'temp_folder' = (string_replace: -find='///', -replace='/', (string_removetrailing: #datetime_folder, -pattern='/'));
									/if;
							
									local: 'sql'='BACKUP TABLE ' + #tables + ' TO "' + #temp_folder + '"';
							
								/if;

								if: (local_defined: 'sql');
									inline: -database=#alias, -sql=#sql;
										if: (error_currenterror) != (error_noerror);
											log_critical: 'Maintenance: Error: ' + (error_currenterror) + ' (' + (error_currenterror: -errorcode) + ')';
										/if;
										records;
											log_warning: '    ' (field: 'table') ': ' + (field: 'msg_type') + ' ' + (field: 'msg_text');									
										/records;
									/inline;
								else;
									log_critical: 'Maintenance: Error: Unknown Error';
									return;
								/if;
							/Protect;
							
							local: 'delay' = (integer: (field: 'delay'));
							local: 'repeat' = (integer: (field: 'delay'));
				
							// Delete the record if it's a one-shot
							If: (#repeat == 0) || (#delay <= 0);
								log_warning: 'Maintenance: Deleting One-Time Event';
								Inline:  -Database='site', -SQL='DELETE FROM _maintenance WHERE id = ' + (Field: 'id'); 
								/Inline;
							Else;
								local: 'next_execute' = (Date: (Field: 'next_execute'), -format='%Q %T');
								if: #delay % 1440 == 0;
									// handle daily, weekly etc repeats specially to make them repeat on the same time of day,
									// regardless of clock changes like daylight savings switch
									// Added by Johan Slve 10/30/2003
									local: 'next_difference'=0;
									// split #next_execute into time and date (use neutral time for the date)
									local: 'next_execute_time'= #next_execute->(format: '%T');
									local: 'next_execute_date'= (date: (#next_execute -> format: '%Q 12:00:00'));
									// add the delay to get the right date (we will disregard the time from this operation)
									#next_execute_date->(add: -minute=#delay);
									// restore the previous time value
									#next_execute = (date: (#next_execute_date -> format: ('%Q ' + #next_execute_time)));
								else;
									// normal rescheduling
									local: 'date_fix' = (Date)->(Difference: (String: (Date)), -MinutesBetween);
									local: 'next_difference' = #next_execute->(Difference: (Date), -MinutesBetween) + #date_fix;
									local: 'next_difference' = #next_difference + #delay - (#next_difference % #delay);
									#next_execute->(add: -minute=#next_difference);
								/if;
								log_warning: 'Maintenance: Rescheduling to ' + (local: 'next_execute');
								Inline:  -Database='site', -SQL='UPDATE _maintenance SET next_execute=\"' + (Date_Format: #next_execute, -format='%Q %T') + '\" WHERE id = ' + (Field: 'id');
								/Inline;
								If: (global: '__Maintenance_Sleep__')->(isa: 'date') && (#next_execute < (global: '__Maintenance_Sleep__'));
									global: '__Maintenance_Sleep__' = #next_execute;
								/if;
							/If;
							
							// 5 Second Delay Between Events
							If: (Loop_Count < Found_Count);
								Sleep: 5000; 
							/If;
							
						/Records;
					/Inline;

					if: ((Global: '__Maintenance_Status__') != 'Pause') && ((Global: '__Maintenance_Status__') != 'Stop');
						Inline: -Database='site', -SQL='SELECT next_execute FROM _maintenance ORDER BY next_execute LIMIT 1';
							If: (Found_Count > 0);
								local: 'next_execute' = (Date: (Field: 'next_execute'), -format='%Q %T');
								If: (global: '__Maintenance_Sleep__')->(isa: 'date') && (#next_execute < (global: '__Maintenance_Sleep__'));
									global: '__Maintenance_Sleep__' = #next_execute;
								/if;
							/if;
						/inline;
						
						If: (global: '__Maintenance_Sleep__')->(isa: 'date') && ((date_add: (date), -minute=2) < (global: '__Maintenance_Sleep__'));
							global: '__Maintenance_Status__' = 'Sleep';
							Log_Warning: 'Maintenance: Sleeping Until ' + (global: '__Maintenance_Sleep__');
						else;
							global: '__Maintenance_Status__' = 'Active';
						/if;
					/if;

				/If;
				
			/Protect;
			
			// 60 Second Delay Between Active Checks
			Sleep: 60000
			; 
		/While;
		
		Log_Warning: 'Maintenance: Stopping Event Queue';	

	}->(AsAsync); 
?>
